from __future__ import unicode_literals

import click
import pkg_resources


@click.group('scaffolding')
def group():
    pass


class Command(object):

    verbosity = 0

    class options(object):
        simulate = False
        interactive = False
        overwrite = True


@group.command()
@click.argument('template')
@click.argument('destination')
def create(template, destination):
    template = [
        scaffold for scaffold in all_scaffolds() if scaffold.name == template
    ][0]
    variables = {
        'project': destination,
        'package': destination,
        'egg': destination,
        'pyramid_version': '1.3.1',
        'pyramid_docs_branch': 'latest',
    }

    template.run(Command, destination, variables)


@group.command('list')
def list_scaffolding():
    for scaffold in all_scaffolds():
        click.echo(scaffold.name + '\t' + scaffold.summary)


def all_scaffolds():
    scaffolds = []
    eps = list(pkg_resources.iter_entry_points('pyramid.scaffold'))
    for entry in eps:
        try:
            scaffold_class = entry.load()
            scaffold = scaffold_class(entry.name)
            scaffolds.append(scaffold)
        except Exception as ex:  # pragma: no cover
            click.echo(
                'Warning: could not load entry point %s (%s: %s)' % (
                    entry.name, ex.__class__.__name__, ex
                )
            )
    return scaffolds



from __future__ import unicode_literals
import inspect

from pyramid.view import view_defaults

from .resources import ModelBackedResource


class Controller(object):
    def __init__(self, context, request):
        super(Controller, self).__init__()
        self.request = request
        self.context = context


class RestController(Controller):
    registry = {}

    @classmethod
    def register(cls, name, **kwargs):

        def wrapped(controller):
            cls.registry[name] = controller
            controller = view_defaults(**kwargs)(controller)

            return controller

        return wrapped

    @classmethod
    def operations(cls):
        base_methods = inspect.getmembers(
            RestController, predicate=inspect.ismethod
        )
        to_exclude = [name for name, _ in base_methods]
        operation_map = {
            'index': ('GET', False),
            'create': ('POST', False),
            'show': ('GET', True),
            'update': (('PUT', 'PATCH'), True),
            'delete': ('DELETE', True),
            'upsert': ('PUT', True),
        }

        methods = inspect.getmembers(cls, predicate=inspect.ismethod)
        views = [
            method_name
            for method_name, _ in methods
            if not method_name.startswith('_')
            and method_name not in to_exclude
        ]

        for method_name in views:
            if method_name not in operation_map:
                continue
            verbs, for_resource = operation_map[method_name]
            if not isinstance(verbs, (tuple, list)):
                verbs = (verbs,)

            yield method_name, verbs, for_resource

    @classmethod
    def scan(cls, config):
        for resource, controller in cls.registry.iteritems():
            for method_name, verbs, resource in controller.operations():
                view_kwargs = dict(
                    view=controller,
                    request_method=verbs,
                    attr=method_name,
                )
                defaults = controller.__view_defaults__
                context = defaults['context']
                view_kwargs.update(defaults)
                # a resource specific view with a model associated to it will
                # use a resource predicate to lookup the object.
                if resource and issubclass(context, ModelBackedResource):
                    view_kwargs['resource'] = context.model_cls
                config.add_view(**view_kwargs)


def includeme(config):
    RestController.scan(config)



from __future__ import unicode_literals
import logging


logger = logging.getLogger(__name__)


class Resource(object):
    endpoint_map = {}

    def __init__(self, request, parent=None, name=None, entity=None, **kwargs):
        self.__name__ = name or 'root'
        self.parent = parent
        self.request = request
        self.entity = entity
        self.attrs = kwargs

    @classmethod
    def nest(cls, key):
        def wrapped(nested_cls):
            if key in cls.endpoint_map:
                logger.warning(
                    '%s already registered with %s', key,
                    cls.endpoint_map[key]
                )
            cls.endpoint_map[key] = nested_cls
            return nested_cls

        return wrapped

    def _lookup_endpoint_from_map(self, key):
        return self.endpoint_map[key]

    def _create_context_instance(self, cls, key, entity=None):
        if cls is not None:
            return cls(
                request=self.request, parent=self, name=key, entity=entity
            )

    def __getitem__(self, key):
        ctx_cls = self._lookup_endpoint_from_map(key)
        return self._create_context_instance(ctx_cls, key)

    def __repr__(self):
        attributes = dict(
            (key, value)
            for key, value in self.__dict__.iteritems()
            if key != '__name__' and value
        )
        return '<{name}({__name__}): {attributes}>'.format(
            attributes=attributes,
            name=self.__class__.__name__,
            **self.__dict__
        )


class ModelBackedResource(Resource):
    model_cls = None

    def lookup(self, key):
        raise NotImplementedError()

    def __getitem__(self, key):
        model = self.lookup(key)
        if not model:
            return super(ModelBackedResource, self).__getitem__(key)
        return self._create_context_instance(type(self), key, entity=model)


class DBModelBackedResource(ModelBackedResource):

    def lookup(self, key):
        return self.model_cls.query.get(key).one()


def includeme(config):
    config.set_root_factory(Resource)



from __future__ import unicode_literals
import inspect

from sqlalchemy.ext.declarative import api, declarative_base
from sqlalchemy.orm.exc import DetachedInstanceError

from thed.db import metadata


class Model(object):

    @classmethod
    def __declare_last__(cls):
        for base_cls in inspect.getmro(cls):
            if '_register_' in base_cls.__dict__:
                base_cls._register_(cls)
        return cls

    def __init__(self, **kwargs):
        """
        Initializes a model by invoking the _declarative_constructor
        in SQLAlchemy. We do this for full control over construction
        of an object
        """
        api._declarative_constructor(self, **kwargs)

    def __repr__(self):
        cols = self.__mapper__.c.keys()
        class_name = self.__class__.__name__
        try:
            items = ', '.join([
                '%s=%s' % (col, repr(getattr(self, col))) for col in cols
            ])
        except DetachedInstanceError:
            items = '<detached>'
        return '%s(%s)' % (class_name, items)


DBModel = declarative_base(
    cls=Model,
    constructor=None,
    metadata=metadata,
)


def init(settings):
    DBModel.query = settings['session']
    return settings



from __future__ import unicode_literals


class ResourcePredicate(object):
    """
    """

    def __init__(self, val, config):
        self.val = val

    def text(self):
        return 'val=%s' % (self.val,)

    phash = text

    def __call__(self, context, request):
        return context.entity and isinstance(context.entity, self.val)



from __future__ import unicode_literals

from sqlalchemy import engine_from_config, MetaData
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker


metadata = MetaData()


def init(settings):
    if 'db.engine' not in settings:
        settings['db.engine'] = (
            engine_from_config(settings, 'sqlalchemy.')
        )

    if 'session' not in settings:
        settings['session'] = scoped_session(sessionmaker(
            bind=settings['db.engine'],
        ))
    return settings



from __future__ import unicode_literals

import os
import re
import setuptools

version = (
    re
    .compile(r".*__version__ = '(.*?)'", re.S)
    .match(open('thed/__init__.py').read())
    .group(1)
)

packages = setuptools.find_packages('.', exclude=('tests', 'tests.*'))

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.txt')) as f:
    README = f.read()

with open(os.path.join(here, 'CHANGES.txt')) as f:
    CHANGES = f.read()

requires = [
    'pyramid==1.5.1',
    'SQLAlchemy==0.9.6',
    'click==2.2',
    'pilo>=0.3.2,<0.4'
]

extras_require = {
    'tests': [
        'nose>=1.3,<1.4',
        'mock>=1.0,<1.1',
        'ipdb',
        'pyflakes',
        'webtest',
        'WSGIProxy2',
    ]
}

scripts = [
    'bin/thed',
]


setuptools.setup(
    name='thed',
    version=version,
    description='thed',
    long_description=README + '\n\n' + CHANGES,
    classifiers=[
        "Programming Language :: Python",
    ],
    author='Balanced',
    author_email='dev+thed@balancedpayments.com',
    packages=packages,
    include_package_data=True,
    zip_safe=False,
    scripts=scripts,
    install_requires=requires,
    extras_require=extras_require,
    tests_require=extras_require['tests'],
    test_suite='nose.collector',
    entry_points="""\
        [pyramid.scaffold]
        api=thed.scaffolds:APITemplate
      """
)



from __future__ import unicode_literals
import itertools

from pyramid.config import Configurator
from pyramid.authorization import ACLAuthorizationPolicy

from .auth import AuthenticationPolicy
from .controllers import Controller, RestController
import predicates, requests, resources, responses
from .requests import Request
from .resources import Resource, DBModelBackedResource, ModelBackedResource
from .responses import Response


class Application(object):

    @classmethod
    def configurator_cls(cls):
        return Configurator

    @classmethod
    def request_cls(cls):
        return Request

    @classmethod
    def authentication_policy(cls):
        return AuthenticationPolicy()

    @classmethod
    def authorization_policy(cls):
        return ACLAuthorizationPolicy()

    tweens = []

    config_includes = []

    @classmethod
    def create(cls, default_settings, hook=None, includes=None, tweens=None,
               **overrides):
        """
        This function returns a Pyramid WSGI application.
        """
        app_settings = default_settings.copy()
        app_settings.update(overrides)

        config = cls.configurator_cls()(
            settings=app_settings,
            request_factory=cls.request_cls(),
            authentication_policy=cls.authentication_policy(),
            authorization_policy=cls.authorization_policy(),
        )

        for tween in itertools.chain(cls.tweens, tweens or []):
            config.add_tween(tween)

        for include in itertools.chain(cls.config_includes, includes or []):
            config.include(include)

        if hook:
            hook(config)

        config.scan()

        return config.make_wsgi_app()
